#  tests for pysnmp-4.4.4-py36_0 (this is a generated file);
print('===== testing package: pysnmp-4.4.4-py36_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
from pysnmp.entity.rfc3413.oneliner import cmdgen
#  --- run_test.py (end) ---

print('===== pysnmp-4.4.4-py36_0 OK =====');
print("import: 'pysnmp'")
import pysnmp

print("import: 'pysnmp.carrier'")
import pysnmp.carrier

print("import: 'pysnmp.carrier.asyncio'")
import pysnmp.carrier.asyncio

print("import: 'pysnmp.carrier.asyncio.dgram'")
import pysnmp.carrier.asyncio.dgram

print("import: 'pysnmp.carrier.asyncore'")
import pysnmp.carrier.asyncore

print("import: 'pysnmp.carrier.asyncore.dgram'")
import pysnmp.carrier.asyncore.dgram

print("import: 'pysnmp.carrier.asynsock'")
import pysnmp.carrier.asynsock

print("import: 'pysnmp.carrier.asynsock.dgram'")
import pysnmp.carrier.asynsock.dgram

print("import: 'pysnmp.carrier.twisted'")
import pysnmp.carrier.twisted

print("import: 'pysnmp.carrier.twisted.dgram'")
import pysnmp.carrier.twisted.dgram

print("import: 'pysnmp.entity'")
import pysnmp.entity

print("import: 'pysnmp.entity.rfc3413'")
import pysnmp.entity.rfc3413

print("import: 'pysnmp.entity.rfc3413.oneliner'")
import pysnmp.entity.rfc3413.oneliner

print("import: 'pysnmp.hlapi'")
import pysnmp.hlapi

print("import: 'pysnmp.hlapi.asyncio'")
import pysnmp.hlapi.asyncio

print("import: 'pysnmp.hlapi.asyncore'")
import pysnmp.hlapi.asyncore

print("import: 'pysnmp.hlapi.asyncore.sync'")
import pysnmp.hlapi.asyncore.sync

print("import: 'pysnmp.hlapi.asyncore.sync.compat'")
import pysnmp.hlapi.asyncore.sync.compat

print("import: 'pysnmp.hlapi.twisted'")
import pysnmp.hlapi.twisted

print("import: 'pysnmp.proto'")
import pysnmp.proto

print("import: 'pysnmp.proto.acmod'")
import pysnmp.proto.acmod

print("import: 'pysnmp.proto.api'")
import pysnmp.proto.api

print("import: 'pysnmp.proto.mpmod'")
import pysnmp.proto.mpmod

print("import: 'pysnmp.proto.proxy'")
import pysnmp.proto.proxy

print("import: 'pysnmp.proto.secmod'")
import pysnmp.proto.secmod

print("import: 'pysnmp.proto.secmod.eso'")
import pysnmp.proto.secmod.eso

print("import: 'pysnmp.proto.secmod.eso.priv'")
import pysnmp.proto.secmod.eso.priv

print("import: 'pysnmp.proto.secmod.rfc3414'")
import pysnmp.proto.secmod.rfc3414

print("import: 'pysnmp.proto.secmod.rfc3414.auth'")
import pysnmp.proto.secmod.rfc3414.auth

print("import: 'pysnmp.proto.secmod.rfc3414.priv'")
import pysnmp.proto.secmod.rfc3414.priv

print("import: 'pysnmp.proto.secmod.rfc3826'")
import pysnmp.proto.secmod.rfc3826

print("import: 'pysnmp.proto.secmod.rfc3826.priv'")
import pysnmp.proto.secmod.rfc3826.priv

print("import: 'pysnmp.proto.secmod.rfc7860'")
import pysnmp.proto.secmod.rfc7860

print("import: 'pysnmp.proto.secmod.rfc7860.auth'")
import pysnmp.proto.secmod.rfc7860.auth

print("import: 'pysnmp.smi'")
import pysnmp.smi

print("import: 'pysnmp.smi.mibs'")
import pysnmp.smi.mibs

print("import: 'pysnmp.smi.mibs.instances'")
import pysnmp.smi.mibs.instances

